#include <Arduino.h>
#include <WiFi.h>
#include <PubSubClient.h>
#include <ArduinoJson.h>
#include <FastLED.h>
#include <SPI.h>
#include <SD.h>

char* config_filename = "/config.txt";

struct wifi_config_struct
{
  char ssid[32];
  char password[16];
  char ip[16];
  char mask[16];
  char gateway[16];
  char dns1[16];
  char dns2[16];
};

wifi_config_struct wifi_config;


// Load configuration file from SD card
void loadconfig(const char* filename) {
  Serial.println(F("Loading configuration file."));
  File file = SD.open(filename);

  StaticJsonDocument<512> jsonBuffer;

  DeserializationError error = deserializeJson(jsonBuffer, file);
  if (error) {
    Serial.print(F("Unable to deserialize JSON config file."));
    Serial.println(error.f_str());
  } else {
      strcpy(wifi_config.ssid, jsonBuffer["wifi"]["ssid"] | "defualtSSID");
      strcpy(wifi_config.password, jsonBuffer["wifi"]["password"] | "");
      strcpy(wifi_config.ip, jsonBuffer["wifi"]["ip"] | "");
      strcpy(wifi_config.mask, jsonBuffer["wifi"]["mask"] | "");
      strcpy(wifi_config.gateway, jsonBuffer["wifi"]["gateway"] | "");
      strcpy(wifi_config.dns1, jsonBuffer["wifi"]["dns1"] | "");
      strcpy(wifi_config.dns2, jsonBuffer["wifi"]["dns2"] | "");
  }

  file.close();

}

// Setup for addressable RGB LED's
#define num_leds 9
#define data_pin 4

CRGB leds[num_leds];

//Initialize WiFi Function
void initWiFi() {

  // Set Static IP address
  IPAddress local_IP;
  local_IP.fromString(wifi_config.ip);

  // Set Gateway IP address
  IPAddress gateway;
  gateway.fromString(wifi_config.gateway);
  IPAddress subnet;
  subnet.fromString(wifi_config.mask);
  IPAddress primaryDNS;   // optional
  primaryDNS.fromString(wifi_config.dns1);
  IPAddress secondaryDNS; // optional
  secondaryDNS.fromString(wifi_config.dns2);
  
  // Configures static IP address
  if (!WiFi.config(local_IP, gateway, subnet, primaryDNS, secondaryDNS)) {
    Serial.println(F("STA Failed to configure"));
  }

  WiFi.mode(WIFI_STA);
  WiFi.begin(wifi_config.ssid, wifi_config.password);
  Serial.print(F("Connecting to WiFi .."));
  while (WiFi.status() != WL_CONNECTED) {
    Serial.print('.');
    delay(1000);
  }
  Serial.println();
  Serial.print(F("IP Address: "));
  Serial.println(WiFi.localIP());
}

// Variable for Power Usage
char mqtt_topic[64] = "";
float mqtt_payload_value = 0;

void printBuffers() {
  Serial.print("MQTT_Topic: ");
  Serial.println(mqtt_topic);
  Serial.print("Value: ");
  Serial.println(mqtt_payload_value);
}

// Callback to process MQTT payload
void callback(char* topic, byte* payload, unsigned int length) {
  Serial.print("Message arrived!");

  StaticJsonDocument<64> doc;
  DeserializationError error = deserializeJson(doc, payload, length);
  if (error) {
    Serial.print(F("deserializeJson() failed: "));
    Serial.println(error.f_str());
    return;
  } else {
      mqtt_payload_value = doc["value"];
      Serial.println(mqtt_payload_value);
      strcpy(mqtt_topic, topic);
  }
  printBuffers();
}

//Configure MQTT
WiFiClient wifi_client;
IPAddress mqtt_server(192, 168, 1, 10);
PubSubClient mqtt_client(mqtt_server, 1883, callback, wifi_client);

// Connect to MQTT Server and subscrib to topics
void reconnect_mqtt() {
  while (!mqtt_client.connected()) {
    Serial.println(F("Attempting MQTT connection"));
    if (mqtt_client.connect("ESP32")) {
      Serial.println("MQTT Connected!");
      mqtt_client.subscribe("zwave/nodeID_7/50/0/value/66049");
    } else {
      Serial.println(F("MQTT Connection Failed. Trying again."));
      delay(250);
    }
  }
}

void setup() {
  // Configure Serial
  Serial.begin(9600);

  delay(1000);

  //Initialize SD card. Pin 5 is slave select for SD card
  Serial.println(F("Initializing SD card"));
  if (!SD.begin(5)) {
    Serial.println(F("Initialization failed!"));
    while (1);
  }
  Serial.println(F("SD card Initialization success!"));

  loadconfig(config_filename);

  SD.end();

  // Setup LED's
  FastLED.addLeds<NEOPIXEL, data_pin>(leds, num_leds);

  // Configure Pins
  pinMode(2, OUTPUT);

  // Connect to WiFi
  initWiFi();

  //Connect to MQTT Server
  if (!mqtt_client.connected()) {
    reconnect_mqtt();
  }

  // Create initial LED Pattern
  for (int i = 0; i < sizeof(leds); i++) {
    int mod = i % 7;
    if ( mod == 0) {
      leds[i] = CRGB::Red;
    } else if (i == 1) {
      leds[i] = CRGB::Orange;
    } else if (i == 2) {
      leds[i] = CRGB::Yellow;
    } else if (i == 3) {
      leds[i] = CRGB::Green;
    } else if (i == 4) {
      leds[i] = CRGB::Blue;
    } else if (i == 5) {
      leds[i] = CRGB::Indigo;
    } else if (i == 6) {
      leds[i] = CRGB::Violet;
    }
  }
  FastLED.show();
}

void ESP_LED_flash() {

  static boolean LED_flag = true;
  static unsigned long LED_timer = millis();

  // Change LED state every 250ms
  if (millis() - LED_timer >= 250) {
    LED_timer = millis();
    if (LED_flag) {
      digitalWrite(2, HIGH);
    } else {
      digitalWrite(2, LOW);
    }
    LED_flag = !LED_flag;
  }
}



void loop() {

  static unsigned int initial_hue = 0;
  const unsigned int hue_diff = 12;

  ESP_LED_flash();

  static unsigned long fastled_timer = millis();

  if (millis() - fastled_timer >= 20) {
    fastled_timer = millis();
    fill_rainbow(leds, num_leds, initial_hue, hue_diff);
    initial_hue += hue_diff;
    Serial.println("Updating LED Strip");
    FastLED.show();
  }
 
  //Connect to MQTT Server
  if (!mqtt_client.connected()) {
    reconnect_mqtt();
  }

  // Process MQTT packets
  mqtt_client.loop();
}